<?php

/**
 * IPv6 Address Functions for PHP
 *
 * Functions to manipulate IPv6 addresses for PHP
 *
 * Copyright (C) 2009 Ray Patrick Soucy
 *
 * LICENSE:
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @package   inet6
 * @author    Ray Soucy <rps@soucy.org>
 * @version   1.0.1
 * @copyright 2009 Ray Patrick Soucy
 * @link      http://www.soucy.org/
 * @license   GNU General Public License version 3 or later
 * @since     File available since Release 1.0.1
 */

// Modified for use with MyIP

 /**
  * Expand an IPv6 Address
  *
  * This will take an IPv6 address written in short form and expand it to include all zeros.
  *
  * @param  string  $addr A valid IPv6 address
  * @return string  The expanded notation IPv6 address
  */
function inet6_expand($addr)
{
  /* Check if is IPv4 mapped address */
  if ( preg_match("/:(([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3}))$/", $addr, $matches))
  {
    // Check the value of the 80-96th bits
    if ( preg_match("/ffff:(([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3}))$/", $addr))
    {
      $s = 'ffff';
    } else
    {
      $s = '0000';
    }

    $ipv4_addr = $matches[1];
    $inet = 0.0;
    if (count($t = explode(".", $ipv4_addr)) != 4) return 0;

    for ($i = 0; $i < 4; $i++) {
      $inet *= 256.0;
      $inet += $t[$i];
    };
    // We now have integer representing ipv4 address, now we need hex
    $p = dechex($inet);

    while (strlen($p) < 8) { $p = '0' . $p; }

    $ipv6_hex = "00000000000000000000" . $s . $p;
    $result = '';
    for ($i = 0; $i < 8; $i++) {
      $result .= substr($ipv6_hex, $i * 4, 4);
      if ($i != 7) $result .= ':';
    } // for
    return $result;
  }

    /* Check if there are segments missing, insert if necessary */
    if (strpos($addr, '::') !== false) {
        $part = explode('::', $addr);
        $part[0] = explode(':', $part[0]);
        $part[1] = explode(':', $part[1]);
        $missing = array();
        for ($i = 0; $i < (8 - (count($part[0]) + count($part[1]))); $i++)
            array_push($missing, '0000');
        $missing = array_merge($part[0], $missing);
        $part = array_merge($missing, $part[1]);
    } else {
        $part = explode(":", $addr);
    } // if .. else
    /* Pad each segment until it has 4 digits */
    foreach ($part as &$p) {
        while (strlen($p) < 4) $p = '0' . $p;
    } // foreach
    unset($p);
    /* Join segments */
    $result = implode(':', $part);
    /* Quick check to make sure the length is as expected */
    if (strlen($result) == 39) {
        return $result;
    } else {
        return false;
    } // if .. else
} // inet6_expand

 /**
  * Compress an IPv6 Address
  *
  * This will take an IPv6 address and rewrite it in short form.
  *
  * @param  string  $addr A valid IPv6 address
  * @return string  The address in short form notation
  */
function inet6_compress($addr)
{
    /* PHP provides a shortcut for this operation */
    $result = inet_ntop(inet_pton($addr));

    return $result;
} // inet6_compress

 /**
  * Generate an IPv6 mask from prefix notation
  *
  * This will convert a prefix to an IPv6 address mask (used for IPv6 math)
  *
  * @param  integer $prefix The prefix size, an integer between 1 and 127 (inclusive)
  * @return string  The IPv6 mask address for the prefix size
  */
function inet6_prefix_to_mask($prefix)
{
    /* Make sure the prefix is a number between 1 and 127 (inclusive) */
    $prefix = intval($prefix);
    if ($prefix < 0 || $prefix > 128) return false;
    $mask = '0b';
    for ($i = 0; $i < $prefix; $i++) $mask .= '1';
    for ($i = strlen($mask) - 2; $i < 128; $i++) $mask .= '0';

    $mask = gmp_strval(gmp_init($mask), 16);
    $result = '';
    for ($i = 0; $i < 8; $i++) {
        $result .= substr($mask, $i * 4, 4);
        if ($i != 7) $result .= ':';
    } // for
    return $result;
} // inet6_prefix_to_mask

 /**
  * Convert an IPv6 address and prefix size to an address range for the network.
  *
  * This will take an IPv6 address and prefix and return the first and last address available for the network.
  *
  * @param  string  $addr A valid IPv6 address
  * @param  integer $prefix The prefix size, an integer between 1 and 127 (inclusive)
  * @return array   An array with two strings containing the start and end address for the IPv6 network
  */
function inet6_to_range($addr, $prefix)
{
    $size = gmp_pow(2, 128 - $prefix);
    $addr = gmp_init('0x' . str_replace(':', '', inet6_expand($addr)));
    $mask = gmp_init('0x' . str_replace(':', '', inet6_prefix_to_mask($prefix)));

    $prefix = gmp_and($addr, $mask);
    $broadcast = gmp_add($prefix, gmp_sub($size, 1));

    $start = gmp_strval($prefix, 16);
    $end = gmp_strval($broadcast, 16);
    while (strlen($start) < 32) $start = '0' . $start;
    while (strlen($end) < 32) $end = '0' . $end;

    $start_result = '';
    $end_result = '';
    for ($i = 0; $i < 8; $i++) {
        $start_result .= substr($start, $i * 4, 4);
        if ($i != 7) $start_result .= ':';
    } // for

    for ($i = 0; $i < 8; $i++) {
        $end_result .= substr($end, $i * 4, 4);
        if ($i != 7) $end_result .= ':';
    } // for

    $result = array(inet6_compress($start_result), inet6_compress($end_result));
    return $result;
} // inet6_to_range

 /**
  * Convert an IPv6 address to two 64-bit integers.
  *
  * This will translate an IPv6 address into two 64-bit integer values for storage in an SQL database.
  *
  * @param  string  $addr A valid IPv6 address
  * @return array   An array with two strings containing the 64-bit interger values
  */
function inet6_to_int64($addr)
{
  $addr = (string) $addr;
    /* Expand the address if necessary */
    if (strlen($addr) != 39) {
        $addr = inet6_expand($addr);
        if ($addr == false) return false;
    } // if
    $addr = str_replace(':', '', $addr);
    $p1 = '0x' . substr($addr, 0, 16);
    $p2 = '0x' . substr($addr, 16);
    $p1 = gmp_init($p1);
    $p2 = gmp_init($p2);
    $result = array(gmp_strval($p1), gmp_strval($p2));
    return $result;
} // inet6_to_int64()

 /**
  * Convert two 64-bit integer values into an IPv6 address
  *
  * This will translate an array of 64-bit integer values back into an IPv6 address
  *
  * @param  array  $val An array containing two strings representing 64-bit integer values
  * @return string An IPv6 address
  */
function int64_to_inet6($val)
{
    /* Make sure input is an array with 2 numerical strings */
    $result = false;
    if ( ! is_array($val) || count($val) != 2) return $result;
    $p1 = gmp_strval(gmp_init($val[0]), 16);
    $p2 = gmp_strval(gmp_init($val[1]), 16);
    while (strlen($p1) < 16) $p1 = '0' . $p1;
    while (strlen($p2) < 16) $p2 = '0' . $p2;
    $addr = $p1 . $p2;
    for ($i = 0; $i < 8; $i++) {
        $result .= substr($addr, $i * 4, 4);
        if ($i != 7) $result .= ':';
    } // for
    return inet6_compress($result);
} // int64_to_inet6()

 /**
  * Adds one to an IPv6 ip address.
  *
  * This will add one to an IPv6 address, useful for getting the next network from a broadcast address
  *
  * @param  string  $addr A valid IPv6 address
  * @param  string  $num Number to add (use negative number to subtract!
  * @return string  A valid IPV6 address
  */
function inet6_add($addr,$num)
{
  $addr = (string) $addr;
    /* Expand the address if necessary */
    if (strlen($addr) != 39) {
        $addr = inet6_expand($addr);
        if ($addr == false) return false;
    } // if
    $addr = str_replace(':', '', $addr);
    $p1 = '0x' . $addr;
    $p1 = gmp_init($p1);
    $addr = gmp_add($p1, $num);
    $addr = gmp_strval($addr, 16);
    while (strlen($addr) < 32) $addr = '0' . $addr;

    $result = '';
    for ($i = 0; $i < 8; $i++) {
        $result .= substr($addr, $i * 4, 4);
        if ($i != 7) $result .= ':';
    } // for
    return inet6_compress($result);

} // inet6_add()


 /**
  * Returns 1 if it's an IPv4 address in IPv6 Format, else 0.
  *
  *  Tests wether an IP is IPv4 address in IPv6 format.
  *
  * @param  string  $ipv6_addr A valid IPv6 address
  * @return integer Result of test
  */
function inet6_is_ipv4($ipv6_addr)
{
  $ipv6_addr = gmp_init('0x' . str_replace(':', '', inet6_expand($ipv6_addr)));
    $mask =      gmp_init('0xffffffffffffffffffffffff00000000');
    $ipv4_addr = gmp_init('0x00000000000000000000ffff00000000');
    $result = gmp_and($ipv6_addr, $mask);
    if (gmp_cmp($result, $ipv4_addr) == 0)
    {
      return 1;
    } else
    {
      return 0;
    }
} // inet6_is_ipv4()

 /**
  * Takes an IPv4 Address and returns IPv6 mapped IP
  *
  *  Takes an IPv4 Address and returns IPv6 mapped IP
  *
  * @param  string  $ipv4_addr A valid IPv4 address
  * @return string  A valid IPv6 Address
  */
function ipv4_to_inet6($ipv4_addr)
{

    return "::ffff:".$ipv4_addr;

} // ipv4_to_inet6()

 /**
  * IPv6 to IPv4
  *
  *  Takes an IPv6 Address and returns IPv4 address if valid, otherwise just returns IPv6 addr (used mostly for display)
  *
  * @param  string  $ipv6_addr A valid IPv6 address
  * @return string  A valid IPv4 or IPv6 Address
  */
function inet6_to_ipv4($ipv6_addr)
{
  if (inet6_is_ipv4($ipv6_addr))
  {
    if ( preg_match("/ffff:(([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3}))$/", $ipv6_addr, $matches))
    {
      $ipv4_addr = $matches[1];
      return $ipv4_addr;
    } else
    {
      return $ipv6_addr;
    }
  } else
  {
    return $ipv6_addr;
  }
} // inet6_to_ipv4()

 /**
  * Tests if it's a valid network address for the given prefix
  *
  *  Returns 1 if the subnet / prefix combo are valid
  *
  * @param  string  $addr A valid IPv6 address
  * @param  integer $prefix Prefix of network
  * @return integer Result of test
  */
function inet6_validate_network($addr, $prefix)
{
  // get subnet network, compare it to addr, if they're the same then we're good.
  $addr = gmp_init('0x' . str_replace(':', '', inet6_expand($addr)));
  $mask = gmp_init('0x' . str_replace(':', '', inet6_prefix_to_mask($prefix)));
  if ((gmp_cmp(gmp_and($addr, $mask), $addr)) == 0)
  {
    return 1;
  } else
  {
    return 0;
  }
} // inet6_validate_network


 /**
  * Returns the largest prefix that can be used in the gap between low and high IP's
  *
  * Given Low and High ip's, calculates the largest prefix within the range.
  *
  * @param  string  $low_ip A valid IPv6 address
  * @param  string  $high_ip  A valid IPv6 Address
  * @return integer Largest Prefix
  */
function inet6_largest_prefix($low_ip, $high_ip)
{

  $low_ip = (string) $low_ip;
  $high_ip = (string) $high_ip;

  /* Expand the address if necessary */
    if (strlen($low_ip) != 39) {
        $low_ip = inet6_expand($low_ip);
        if ($low_ip == false) return false;
    } // if
  /* Expand the address if necessary */
    if (strlen($high_ip) != 39) {
        $high_ip = inet6_expand($high_ip);
        if ($high_ip == false) return false;
    } // if

  $low_ip = str_replace(':', '', $low_ip);
    $p1 = '0x' . $low_ip;
    $p1 = gmp_init($p1);

    $high_ip = str_replace(':', '', $high_ip);
    $p2 = '0x' . $high_ip;
    $p2 = gmp_init($p2);

    $size = gmp_sub($p2, $p1);

    $i = 0;
    while ( $i <= 128)
    {
      $s = gmp_pow(2, 128 - $i);
      if (gmp_cmp($size, $s) > 0)
      {
        return $i;
      }
      $i++;
    }

} // inet6_largest_prefix()